import pydantic

from huma_signals.commons import chains


class Settings(pydantic.BaseSettings):
    class Config:
        case_sensitive = False

    chain: chains.Chain
    superfluid_subgraph_endpoint_url: str


settings = Settings()



import decimal
import pathlib
from typing import Any, ClassVar

import aiofiles
import orjson
import pydantic
import structlog
import web3
from web3 import exceptions as web3_exceptions

from huma_signals import exceptions, models
from huma_signals.adapters import models as adapter_models
from huma_signals.adapters.lending_pools import registry
from huma_signals.adapters.lending_pools.settings import settings
from huma_signals.commons import web3_utils

logger = structlog.get_logger(__name__)


class LendingPoolSignals(models.HumaBaseModel):
    # TODO: add other pool signals: utilization, liquidity, etc.
    # Pool policy signals
    pool_address: str = pydantic.Field(..., description="Address of the lending pool")
    apr: int = pydantic.Field(..., description="Annual percentage rate in BPS")
    max_credit_amount: decimal.Decimal = pydantic.Field(
        ..., description="Maximum credit amount defined by the pool"
    )
    token_name: str = pydantic.Field(
        ..., description="Name of the token used by the pool"
    )
    token_symbol: str = pydantic.Field(
        ..., description="Symbol of the token used by the pool"
    )
    token_decimal: int = pydantic.Field(
        ..., description="Decimal of the token used by the pool"
    )
    interval_in_days_max: int = pydantic.Field(
        ..., description="Maximum payment interval in days for the pool"
    )
    interval_in_days_min: int = pydantic.Field(
        ..., description="Minimum payment interval in days for the pool"
    )
    invoice_amount_ratio: float = pydantic.Field(
        ...,
        description="For invoice factoring pool, the percentage of the payable amount the pool allow to borrow",
    )
    is_testnet: bool = pydantic.Field(
        default=False, description="Whether the pool is on testnet or not"
    )


class LendingPoolAdapter(adapter_models.SignalAdapterBase):
    # TODO: move the hard coded values to EA settings cause these are not pool setting
    interval_in_days_max: ClassVar[int] = 90
    interval_in_days_min: ClassVar[int] = 0
    invoice_amount_ratio: ClassVar[float] = 0.8

    async def fetch(  # pylint: disable=arguments-differ
        self, pool_address: str, *args: Any, **kwargs: Any
    ) -> LendingPoolSignals:
        try:
            checksum_address = web3.Web3.to_checksum_address(pool_address)
            pool_settings = registry.POOL_REGISTRY[checksum_address]
        except KeyError as e:
            message = f"Invalid pool_address {pool_address}: pool settings not found."
            logger.exception(message)
            raise exceptions.PoolSettingsNotFoundException(
                pool_address=pool_address
            ) from e

        w3 = await web3_utils.get_w3(pool_settings.chain, settings.web3_provider_url)

        async with aiofiles.open(pool_settings.pool_abi_path) as f:
            contents = await f.read()
            huma_pool_contract = w3.eth.contract(
                address=checksum_address,
                abi=orjson.loads(contents),
            )
        try:
            contract_address = await huma_pool_contract.functions.poolConfig().call()
        except web3_exceptions.Web3Exception as e:
            message = "Failed to get contract address"
            logger.exception(message)
            raise exceptions.ContractCallFailedException(message=message) from e

        async with aiofiles.open(
            pathlib.Path(__file__).parent.resolve() / "abi" / "BasePoolConfig.json",
        ) as f:
            contents = await f.read()
            pool_config_contract = w3.eth.contract(
                address=contract_address,
                abi=orjson.loads(contents),
            )

        try:
            pool_summary = await pool_config_contract.functions.getPoolSummary().call()
        except web3_exceptions.Web3Exception as e:
            message = "Failed to get pool summary"
            logger.exception(message)
            raise exceptions.ContractCallFailedException(message=message) from e

        return LendingPoolSignals(
            pool_address=pool_address,
            apr=pool_summary[1],
            max_credit_amount=pool_summary[3],
            token_name=pool_summary[5],
            token_symbol=pool_summary[6],
            token_decimal=pool_summary[7],
            interval_in_days_max=self.interval_in_days_max,
            interval_in_days_min=self.interval_in_days_min,
            invoice_amount_ratio=self.invoice_amount_ratio,
            is_testnet=pool_settings.chain.is_testnet(),
        )



import httpx
import structlog
import web3

from huma_signals import exceptions
from huma_signals.adapters import models as adapter_models
from huma_signals.adapters.superfluid import superfluid_models
from huma_signals.adapters.superfluid.settings import settings
from huma_signals.commons import chains

logger = structlog.get_logger()


_CURRENT_STREAM_QUERY = """
query HumaCurrentStream($sender: String, $receiver: String, $token: String) {
    streams(
        where: {
            sender: $sender,
            receiver: $receiver,
            token: $token
        }
        first: 1
        orderBy: updatedAtTimestamp
        orderDirection: desc
        currentFlowRate_gt: 0
    ) {
        id
        currentFlowRate
        createdAtTimestamp
        updatedAtTimestamp
    }
}
"""


class SuperfluidAdapter(adapter_models.SignalAdapterBase):
    def __init__(
        self,
        superfluid_subgraph_endpoint_url: str = settings.superfluid_subgraph_endpoint_url,
        chain: chains.Chain = settings.chain,
    ) -> None:
        self.superfluid_subgraph_endpoint_url = superfluid_subgraph_endpoint_url
        self.chain = chain

    async def fetch(  # pylint: disable=arguments-differ
        self,
        borrower_wallet_address: str,
        payer_wallet_address: str,
        super_token_address: str,
    ) -> superfluid_models.SuperfluidSignals:
        for address in [
            borrower_wallet_address,
            payer_wallet_address,
            super_token_address,
        ]:
            if not web3.Web3.is_address(address):
                raise exceptions.InvalidAddressException(f"Invalid address: {address}")

        current_stream = await self._get_current_stream(
            sender_address=payer_wallet_address.lower(),
            receiver_address=borrower_wallet_address.lower(),
            token_address=super_token_address.lower(),
        )
        return superfluid_models.SuperfluidSignals(
            current_flow_rate=current_stream.current_flow_rate,
        )

    async def _get_current_stream(
        self, sender_address: str, receiver_address: str, token_address: str
    ) -> superfluid_models.SuperfluidStream:
        try:
            async with httpx.AsyncClient() as client:
                resp = await client.post(
                    self.superfluid_subgraph_endpoint_url,
                    json={
                        "query": _CURRENT_STREAM_QUERY,
                        "variables": {
                            "sender": sender_address,
                            "receiver": receiver_address,
                            "token": token_address,
                        },
                    },
                )
                streams = resp.json()["data"]["streams"]
                return superfluid_models.SuperfluidStream(**streams[0])
        except KeyError as e:
            message = "No data returned from query"
            logger.exception(message, resp_body=resp.json())
            raise exceptions.SuperfluidException(message=message) from e
        except IndexError as e:
            message = (
                "Stream not found for sender, receiver and token:"
                f" ({sender_address}, {receiver_address}, {token_address})"
            )
            logger.exception(message)
            raise exceptions.SuperfluidStreamNotFoundException(message=message) from e
        except Exception as e:
            message = f"Error fetching Superfluid streams: {e}"
            logger.exception(message)
            raise exceptions.SuperfluidException(message=message) from e



import pydantic

from huma_signals import models
from huma_signals.commons import pydantic_utils


class SuperfluidSignals(models.HumaBaseModel):
    current_flow_rate: int = pydantic.Field(
        description="The current flow rate of the stream in the unit of wei/sec"
    )


class SuperfluidStream(pydantic_utils.CamelCaseAliased):
    id: str
    updated_at_timestamp: int
    created_at_timestamp: int
    current_flow_rate: int



# pylint: disable=too-few-public-methods

import pydantic


class HumaBaseModel(pydantic.BaseModel):
    class Config:
        validate_assignment = True
        arbitrary_types_allowed = True
        anystr_strip_whitespace = True
        allow_population_by_field_name = True
        underscore_attrs_are_private = True



import pathlib

import eth_typing
import web3

from huma_signals import models
from huma_signals.commons import chains


class PoolSetting(models.HumaBaseModel):
    pool_address: eth_typing.ChecksumAddress
    chain: chains.Chain
    pool_abi_path: str


_POOLS = [
    PoolSetting(
        pool_address=web3.Web3.to_checksum_address(
            "0xA22D20FB0c9980fb96A9B0B5679C061aeAf5dDE4"
        ),
        chain=chains.Chain.GOERLI,
        pool_abi_path=str(
            pathlib.Path(__file__).parent.resolve() / "abi" / "BaseCreditPool.json"
        ),
    ),
    PoolSetting(
        pool_address=web3.Web3.to_checksum_address(
            "0x11672c0bBFF498c72BC2200f42461c0414855042"
        ),
        chain=chains.Chain.GOERLI,
        pool_abi_path=str(
            pathlib.Path(__file__).parent.resolve()
            / "abi"
            / "ReceivableFactoringPool.json"
        ),
    ),
    PoolSetting(
        pool_address=web3.Web3.to_checksum_address(
            "0xC5BF9750A7BF93479990EF940d7e3984caa22558"
        ),
        chain=chains.Chain.MUMBAI,
        pool_abi_path=str(
            pathlib.Path(__file__).parent.resolve()
            / "abi"
            / "SuperfluidFactoringPool.json"
        ),
    ),
    PoolSetting(
        pool_address=web3.Web3.to_checksum_address(
            "0xAb3dc5221F373Dd879BEc070058c775A0f6Af759"
        ),
        chain=chains.Chain.POLYGON,
        pool_abi_path=str(
            pathlib.Path(__file__).parent.resolve() / "abi" / "BaseCreditPool.json"
        ),
    ),
    PoolSetting(
        pool_address=web3.Web3.to_checksum_address(
            "0x58AAF1f9cB10F335111A2129273056bbED251B61"
        ),
        chain=chains.Chain.POLYGON,
        pool_abi_path=str(
            pathlib.Path(__file__).parent.resolve()
            / "abi"
            / "ReceivableFactoringPool.json"
        ),
    ),
]


POOL_REGISTRY = {pool.pool_address: pool for pool in _POOLS}



import datetime


def timestamp_to_tz_aware_utc_datetime(timestamp: int | str) -> datetime.datetime:
    tz_naive_dt = datetime.datetime.fromtimestamp(int(timestamp))
    return tz_naive_dt.astimezone(datetime.timezone.utc)


def tz_aware_utc_now() -> datetime.datetime:
    return datetime.datetime.now(tz=datetime.timezone.utc)



def snake_to_camel(name: str, overrides: dict[str, str] | None = None) -> str:
    """
    Converts name from snake case to camel case. If name is in overrides, then return
    the overridden value instead.

    Examples:

    assert snake_to_camel("foo_bar_baz") == "fooBarBaz"
    assert snake_to_camel("foo_bar_baz", {"foo_bar_baz": "fooBARBaz}) == "fooBARBaz"
    """
    if overrides is not None and name in overrides:
        return overrides[name]

    words = [word.title() for word in name.split("_")]
    words[0] = words[0].lower()
    return "".join(words)



import enum


class Chain(enum.Enum):
    ETHEREUM = "ETHEREUM"
    GOERLI = "GOERLI"
    POLYGON = "POLYGON"
    MUMBAI = "MUMBAI"

    def chain_name(self) -> str:
        return self.name.lower()

    def is_testnet(self) -> bool:
        return self.chain_name() in ("goerli", "mumbai")



from typing import Any

import pandas as pd
import structlog
import web3

from huma_signals import exceptions
from huma_signals.adapters import models as adapter_models
from huma_signals.adapters.ethereum_wallet import adapter as ethereum_wallet_adapter
from huma_signals.adapters.polygon_wallet import adapter as polygon_wallet_adapter
from huma_signals.adapters.request_network import models
from huma_signals.adapters.request_network.settings import settings
from huma_signals.clients.request_client import request_client
from huma_signals.commons import chains

logger = structlog.get_logger(__name__)

_WALLET_ADAPTER_BY_CHAIN = {
    chains.Chain.ETHEREUM: ethereum_wallet_adapter.EthereumWalletAdapter,
    chains.Chain.GOERLI: ethereum_wallet_adapter.EthereumWalletAdapter,
    chains.Chain.POLYGON: polygon_wallet_adapter.PolygonWalletAdapter,
}


class RequestTransactionAdapter(adapter_models.SignalAdapterBase):
    def __init__(  # pylint: disable=too-many-arguments
        self,
        request_client_: request_client.BaseRequestClient | None = None,
        wallet_adapter: ethereum_wallet_adapter.BaseEthereumWalletAdapter
        | polygon_wallet_adapter.BasePolygonWalletAdapter
        | None = None,
        request_network_subgraph_endpoint_url: str = settings.request_network_subgraph_endpoint_url,
        invoice_api_url: str = settings.request_network_invoice_api_url,
        chain: chains.Chain = settings.chain,
    ) -> None:
        self.request_client = request_client_ or request_client.RequestClient(
            request_network_subgraph_endpoint_url=request_network_subgraph_endpoint_url,
            invoice_api_url=invoice_api_url,
        )
        self.chain = chain
        if wallet_adapter is not None:
            self.wallet_adapter = wallet_adapter
        else:
            try:
                self.wallet_adapter = _WALLET_ADAPTER_BY_CHAIN[self.chain]()
            except KeyError as e:
                raise ValueError(
                    f"Unsupported chain for wallet tenure: {self.chain}"
                ) from e

    async def fetch(  # pylint: disable=arguments-differ
        self,
        payer_address: str,
        payee_address: str,
        *args: Any,
        **kwargs: Any,
    ) -> models.RequestTransactionSignals:
        if not web3.Web3.is_address(payer_address):
            raise exceptions.InvalidAddressException(
                f"Invalid payer address: {payer_address}"
            )
        if not web3.Web3.is_address(payee_address):
            raise exceptions.InvalidAddressException(
                f"Invalid payee address: {payee_address}"
            )

        payer_payments = await self.request_client.get_payments(
            from_address=payer_address,
            to_address=None,
        )
        payee_payments = await self.request_client.get_payments(
            from_address=None,
            to_address=payee_address,
        )
        payments_df = pd.DataFrame.from_records([*payer_payments, *payee_payments])
        enriched_payments_df = self.request_client.enrich_payments_data(
            payments_df, self.chain
        )

        payer_stats = self.request_client.get_payment_stats(
            enriched_payments_df[enriched_payments_df["from"] == payer_address]
        )
        payee_stats = self.request_client.get_payment_stats(
            enriched_payments_df[enriched_payments_df["to"] == payee_address]
        )
        pair_stats = self.request_client.get_payment_stats(
            enriched_payments_df[
                (enriched_payments_df["from"] == payer_address)
                & (enriched_payments_df["to"] == payee_address)
            ]
        )

        payee_wallet: ethereum_wallet_adapter.EthereumWalletSignals | polygon_wallet_adapter.PolygonWalletSignals
        payer_wallet: ethereum_wallet_adapter.EthereumWalletSignals | polygon_wallet_adapter.PolygonWalletSignals
        payee_wallet = await self.wallet_adapter.fetch(payee_address)
        payer_wallet = await self.wallet_adapter.fetch(payer_address)

        return models.RequestTransactionSignals(
            payer_tenure=payer_wallet.wallet_tenure_in_days,
            payer_recent=int(payer_stats.get("last_txn_age_in_days", 0)),
            payer_count=int(payer_stats.get("total_txns", 0)),
            payer_total_amount=int(payer_stats.get("total_amount", 0)),
            payer_unique_payees=int(payer_stats.get("unique_payees", 0)),
            payee_tenure=payee_wallet.wallet_tenure_in_days,
            payee_recent=int(payee_stats.get("last_txn_age_in_days", 0)),
            payee_count=int(payee_stats.get("total_txns", 0)),
            payee_total_amount=int(payee_stats.get("total_amount", 0)),
            payee_unique_payers=int(payee_stats.get("unique_payers", 0)),
            mutual_count=int(pair_stats.get("total_txns", 0)),
            mutual_total_amount=int(pair_stats.get("total_amount", 0)),
        )



from __future__ import annotations

import decimal

import pydantic
import structlog

from huma_signals import models

logger = structlog.get_logger()


class RequestInvoiceSignals(models.HumaBaseModel):
    # transactions based features: Payer Quality
    payer_tenure: int = pydantic.Field(
        ..., description="The number of days since the payer's first transaction"
    )
    payer_recent: int = pydantic.Field(
        ..., description="The number of days since the payer's most recent transaction"
    )
    payer_count: int = pydantic.Field(
        ..., description="The number of transactions the payer has made"
    )
    payer_total_amount: int = pydantic.Field(
        ..., description="The total amount the payer has paid"
    )
    payer_unique_payees: int = pydantic.Field(
        ..., description="The number of unique payees the payer has paid"
    )

    # transactions based features: Payee Quality
    payee_tenure: int = pydantic.Field(
        ..., description="The number of days since the payee's first transaction"
    )
    payee_recent: int = pydantic.Field(
        ..., description="The number of days since the payee's most recent transaction"
    )
    payee_count: int = pydantic.Field(
        ..., description="The number of transactions the payee has made"
    )
    payee_total_amount: int = pydantic.Field(
        ..., description="The total amount the payee has received"
    )
    payee_unique_payers: int = pydantic.Field(
        ..., description="The number of unique payers the payee has received"
    )

    # transactions based features: Pair Quality
    mutual_count: int = pydantic.Field(
        ..., description="The number of transactions between the payer and payee"
    )
    mutual_total_amount: int = pydantic.Field(
        ..., description="The total amount the payer and payee have transacted"
    )

    # invoice based features
    payee_match_borrower: bool = pydantic.Field(
        ..., description="Whether the borrower is the invoice's payee"
    )
    payer_match_payee: bool = pydantic.Field(
        ..., description="Whether the payee is the invoice's payer"
    )
    borrower_own_invoice: bool = pydantic.Field(
        ..., description="Whether the borrower own the invoice NFT token"
    )
    days_until_due_date: int = pydantic.Field(
        ..., description="The number of days until the invoice's due date"
    )
    invoice_amount: decimal.Decimal = pydantic.Field(
        ..., description="The amount of the invoice"
    )

    # allowlist feature
    payer_on_allowlist: bool = pydantic.Field(
        ..., description="Whether the payer is on the allowlist"
    )


class RequestTransactionSignals(models.HumaBaseModel):
    # transactions based features: Payer Quality
    payer_tenure: int = pydantic.Field(
        ..., description="The number of days since the payer's first transaction"
    )
    payer_recent: int = pydantic.Field(
        ..., description="The number of days since the payer's most recent transaction"
    )
    payer_count: int = pydantic.Field(
        ..., description="The number of transactions the payer has made"
    )
    payer_total_amount: int = pydantic.Field(
        ..., description="The total amount the payer has paid"
    )
    payer_unique_payees: int = pydantic.Field(
        ..., description="The number of unique payees the payer has paid"
    )

    # transactions based features: Payee Quality
    payee_tenure: int = pydantic.Field(
        ..., description="The number of days since the payee's first transaction"
    )
    payee_recent: int = pydantic.Field(
        ..., description="The number of days since the payee's most recent transaction"
    )
    payee_count: int = pydantic.Field(
        ..., description="The number of transactions the payee has made"
    )
    payee_total_amount: int = pydantic.Field(
        ..., description="The total amount the payee has received"
    )
    payee_unique_payers: int = pydantic.Field(
        ..., description="The number of unique payers the payee has received"
    )

    # transactions based features: Pair Quality
    mutual_count: int = pydantic.Field(
        ..., description="The number of transactions between the payer and payee"
    )
    mutual_total_amount: int = pydantic.Field(
        ..., description="The total amount the payer and payee have transacted"
    )



import datetime
from typing import Any

import pandas as pd
import structlog
import web3

from huma_signals import exceptions
from huma_signals.adapters import models as adapter_models
from huma_signals.adapters.ethereum_wallet import adapter as ethereum_wallet_adapter
from huma_signals.adapters.polygon_wallet import adapter as polygon_wallet_adapter
from huma_signals.adapters.request_network import models
from huma_signals.adapters.request_network.settings import settings
from huma_signals.clients.request_client import request_client
from huma_signals.commons import chains

logger = structlog.get_logger(__name__)

_WALLET_ADAPTER_BY_CHAIN = {
    chains.Chain.ETHEREUM: ethereum_wallet_adapter.EthereumWalletAdapter,
    chains.Chain.GOERLI: ethereum_wallet_adapter.EthereumWalletAdapter,
    chains.Chain.POLYGON: polygon_wallet_adapter.PolygonWalletAdapter,
}


class RequestInvoiceAdapter(adapter_models.SignalAdapterBase):
    def __init__(  # pylint: disable=too-many-arguments
        self,
        request_client_: request_client.BaseRequestClient | None = None,
        wallet_adapter: ethereum_wallet_adapter.BaseEthereumWalletAdapter
        | polygon_wallet_adapter.BasePolygonWalletAdapter
        | None = None,
        request_network_subgraph_endpoint_url: str = settings.request_network_subgraph_endpoint_url,
        invoice_api_url: str = settings.request_network_invoice_api_url,
        chain: chains.Chain = settings.chain,
    ) -> None:
        self.request_client = request_client_ or request_client.RequestClient(
            request_network_subgraph_endpoint_url=request_network_subgraph_endpoint_url,
            invoice_api_url=invoice_api_url,
        )
        self.chain = chain
        if wallet_adapter is not None:
            self.wallet_adapter = wallet_adapter
        else:
            try:
                self.wallet_adapter = _WALLET_ADAPTER_BY_CHAIN[self.chain]()
            except KeyError as e:
                raise ValueError(
                    f"Unsupported chain for wallet tenure: {self.chain}"
                ) from e

    async def fetch(  # pylint: disable=too-many-arguments, arguments-differ
        self,
        borrower_wallet_address: str,
        receivable_param: str,
        *args: Any,
        **kwargs: Any,
    ) -> models.RequestInvoiceSignals:
        if not web3.Web3.is_address(borrower_wallet_address):
            raise exceptions.InvalidAddressException(
                f"Invalid borrower wallet address: {borrower_wallet_address}"
            )

        invoice = await self.request_client.get_invoice(invoice_id=receivable_param)
        records = []
        records.extend(
            await self.request_client.get_payments(
                from_address=invoice.payer, to_address=None
            )
        )
        records.extend(
            await self.request_client.get_payments(
                from_address=None, to_address=invoice.payee
            )
        )
        payments_df = pd.DataFrame.from_records(records)
        enriched_payments_df = self.request_client.enrich_payments_data(
            payments_df, chain=self.chain
        )

        payer_stats = self.request_client.get_payment_stats(
            enriched_payments_df[enriched_payments_df["from"] == invoice.payer]
        )
        payee_stats = self.request_client.get_payment_stats(
            enriched_payments_df[enriched_payments_df["to"] == invoice.payee]
        )
        pair_stats = self.request_client.get_payment_stats(
            enriched_payments_df[
                (enriched_payments_df["from"] == invoice.payer)
                & (enriched_payments_df["to"] == invoice.payee)
            ]
        )

        # Fetch wallet tenure
        payee_wallet: ethereum_wallet_adapter.EthereumWalletSignals | polygon_wallet_adapter.PolygonWalletSignals
        payer_wallet: ethereum_wallet_adapter.EthereumWalletSignals | polygon_wallet_adapter.PolygonWalletSignals
        payee_wallet = await self.wallet_adapter.fetch(invoice.payee)
        payer_wallet = await self.wallet_adapter.fetch(invoice.payer)

        return models.RequestInvoiceSignals(
            payer_tenure=payer_wallet.wallet_tenure_in_days,
            payer_recent=int(payer_stats.get("last_txn_age_in_days", 0)),
            payer_count=int(payer_stats.get("total_txns", 0)),
            payer_total_amount=int(payer_stats.get("total_amount", 0)),
            payer_unique_payees=int(payer_stats.get("unique_payees", 0)),
            payee_tenure=payee_wallet.wallet_tenure_in_days,
            payee_recent=int(payee_stats.get("last_txn_age_in_days", 0)),
            payee_count=int(payee_stats.get("total_txns", 0)),
            payee_total_amount=int(payee_stats.get("total_amount", 0)),
            payee_unique_payers=int(payee_stats.get("unique_payers", 0)),
            mutual_count=int(pair_stats.get("total_txns", 0)),
            mutual_total_amount=int(pair_stats.get("total_amount", 0)),
            payee_match_borrower=(
                invoice.payee.lower() == borrower_wallet_address.lower()
            ),
            borrower_own_invoice=(
                invoice.token_owner.lower() == borrower_wallet_address.lower()
            ),
            payer_match_payee=(invoice.payer.lower() == invoice.payee.lower()),
            days_until_due_date=(invoice.due_date - datetime.datetime.utcnow()).days,
            invoice_amount=invoice.amount,
            payer_on_allowlist=True,
        )
